"""We use the acronym SBERT as a catch-all for BERT-based sentence transformers. In particular, we use a lightweight/fast version of one the top-performing model.

Links:
    sbert: https://www.sbert.net/docs/pretrained_models.html#sentence-embedding-models.
    HF: https://huggingface.co/sentence-transformers
"""

import torch
import numpy as np
from .vectorizer import Vectorizer
from tqdm import tqdm

from sentence_transformers import SentenceTransformer

MPS_DEVICE = torch.device("mps")

# MODEL_PATH = "bert-base-nli-mean-tokens" # NOTE: while Liu and Xu (2020) use this model in a metascience context, huggingface recommends a more recent sentence transformer.
MODEL_PATH = "all-MiniLM-L6-v2"  # All-round model tuned for many use-cases. Trained on a large and diverse dataset of over 1 billion training pairs. Listed as rank 50 on https://huggingface.co/spaces/mteb/leaderboard on 10/11/2023 with an average of 56; rank 1 achieved 64, bert-base-uncased achieved 34; GPT embedding ada-002 achieved 60.
EMBEDDING_DIM = 384
MAX_SEQ_LENGTH = 256

BATCH_SIZE = 64


class SBERTVectorizer(Vectorizer):
    def __init__(self, device="cuda", **kwargs) -> None:
        # Get the model
        self.model = SentenceTransformer(MODEL_PATH)

        # set device to GPU
        if device == "mps":
            self.device = MPS_DEVICE
        elif device == "cuda":
            self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

        print(f"Using device: {self.device}.")
        self.model.to(self.device)

        # Put the model in "evaluation" mode
        self.model.eval()
        super().__init__()

    def embed_documents(
        self, docs: list[str], batch_size: int = BATCH_SIZE
    ) -> np.ndarray:
        """Embed a list of documents (raw text) into SBERT vectors, by batching.

        Args:
            docs: the documents to embed.

        Returns:
            a numpy array of shape `(num_documents, 384)`
        """
        if batch_size is None:
            batch_size = BATCH_SIZE

        embeddings = []

        pbar = tqdm(
            total=len(docs),
            desc="embedding documents",
            leave=True,
        )

        for i in range(0, len(docs), batch_size):
            batch = docs[i : min(len(docs), i + batch_size)]

            # no need to convert anything or dig inside model for outputs
            batched_embeddings = self.model.encode(batch)

            # Collect batched embeddings
            embeddings.extend(batched_embeddings)

            pbar.update(batch_size)
        pbar.close()

        # We don't have to deal with OOV, so we always return full list of ids
        return {
            "embeddings": np.array(embeddings),
            "success_indices": np.arange(len(embeddings), dtype=int),
            "fail_indices": np.array([], dtype=int),
        }



from ..mapping.publication import Publication

from abc import ABC, abstractmethod
from functools import partial
from typing import Any
from multiprocessing import Pool
from tqdm import tqdm


class Librarian(ABC):
    @abstractmethod
    def bibtex_entry_identifier(self, bibtex_entry: dict) -> str:
        """Parse a bibtex entry for a usable unique identifier appropriate to the API."""
        raise NotImplementedError

    @abstractmethod
    def get_publications(
        self,
        identifiers: list[str],
        *args,
        call_size: int = None,
        n_attempts_per_query: int = None,
        convert: bool = True,
        **kwargs,
    ) -> list[Publication]:
        """Call an API and retrieve the publications corresponding to str identifiers.

        Args:
            n_attempts_per_query: Number of attempts to access the API per query. Useful when experiencing connection issues.

            call_size: (int): maximum number of papers to call API for in one query; if less than `len(paper_ids)`, chunking will be performed.
        """
        raise NotImplementedError

    @abstractmethod
    def convert_publication(self, pub: Any, *args, **kwargs):
        """Convert an API-specific resulting publication data structure into a sciterra Publication object."""
        raise NotImplementedError

    def convert_publications(
        self,
        papers: list,
        *args,
        multiprocess: bool = True,
        num_processes=6,
        **kwargs,
    ) -> list[Publication]:
        """Convet a list of API-specific results to sciterra Publications, possibly using multiprocessing."""

        # TODO: you need to pass args and kwargs into these
        if not multiprocess:
            return [
                self.convert_publication(
                    paper,
                    *args,
                    **kwargs,
                )
                for paper in papers
            ]
        with Pool(processes=num_processes) as p:
            publications = list(
                tqdm(
                    p.map(
                        partial(self.convert_publication, *args, **kwargs),
                        papers,
                        chunksize=10,
                    ),
                    total=len(papers),
                )
            )

        return publications



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

requirements = [
    "ads",
    "bibtexparser",
    "gensim",  # N.B.: requires scipy <= 1.10.1
    "numpy",
    "pandas",
    "plotnine",
    "scikit-learn",
    "scipy==1.10.1",
    "sentence-transformers",
    "semanticscholar==0.5.0",  # 0.6.0 has AsyncSemanticScholar which creates difficulties
    "spacy",
    "torch",
    "transformers",
]

test_requirements = [
    "black",
    "coverage",
    "pytest",
]

setuptools.setup(
    name="sciterra",
    version="0.0.21",
    author="Nathaniel Imel",
    author_email="nimel@uci.edu",
    description="Scientific literature data exploration analysis",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/nathimel/sciterra",
    project_urls={"Bug Tracker": "https://github.com/nathimel/sciterra/issues"},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    install_requires=requirements,
    extra_requires={"test": test_requirements},
    #    python_requires=">=3.10.6",  # Colab-compatible
    #    python_requires=">=3.9.1",  # cc-compatible
)



"""Test the basic publication wrapper."""

import pytest

from datetime import datetime
from sciterra.mapping.publication import Publication


single_pub_bibtex_fp = "src/tests/data/single_publication.bib"


class TestPublication:
    def test_empty_publication(self):
        # Publication must have data
        with pytest.raises(TypeError) as e_info:
            pub = Publication()

    def test_dummy_publication(self):
        data = {
            "identifier": "exampleidentifierstring",
            "abstract": "Example abstract text.",
            "publication_date": datetime.today().date(),
            "citation_count": 0,
            "url": "exampleurl.com",
            "extra": Publication,  # extra garbage
        }
        pub = Publication(data)

        assert pub.identifier == "exampleidentifierstring"
        assert str(pub) == pub.identifier
        assert pub.abstract == "Example abstract text."
        assert pub.publication_date == datetime.today().date()
        assert pub.citation_count == 0
        assert pub.url == "exampleurl.com"
        assert not hasattr(pub, "extra")
